%% mall_acyclicity 
% 
% Checks whether a list of premises is acyclic with respect to a given metavariable.
% Furthermore, transforms the cyclic premises into acyclic by applying weakening. Note, 
% that this transformation is valid only if the weakinging is present in the base calculus. 

%% is_cyclic_on_mv(+Premises, +Mv)
% 
% Checks whether the given list of premises is acyclic with respect
% to the given metavariavle. 
%
% +Premises 	- list of the premises to be checked
% +Mv		- a negation free metavariavble in the premises. 
%
% covers the case where there is one premise having multiple occurrences
% of Mv and another one havingmultiple occurrences of -Mv
is_cyclic_on_mv(Premises, Mv) :-
	count_having_multi_occurrence(Premises, Mv, R1),
	count_having_multi_occurrence(Premises, -Mv, R2),
	R1 > 0, 
	R2 > 0.

% covers the case where there is one premise having at least one occurrence
% of Mv and one occurrence of -Mv.
is_cyclic_on_mv(Premises, Mv) :-
	count_having_mixed_occurrence(Premises, Mv, R1),
	R1 > 0.

%% convert_to_acyclic(+CyclicPremises, +Mv, -AcyclicPremises)
% 
% Converts the given list of cyclic premises to acyclic with respect 
% to the given metavariable. 
% 
% + Premises 	- list of cyclic premises w.r.t Mv
% + Mv			- a negation free meta variable. 
% - AcPremises	- a list of acyclic premises w.r.t Mv
convert_to_acyclic(Premises, Mv, AcPremises) :-
	is_cyclic_on_mv(Premises, Mv),
	count_having_multi_occurrence(Premises, Mv, R1),
	count_having_multi_occurrence(Premises, -Mv, R2),
	count_having_mixed_occurrence(Premises, Mv, R3),
	neg_free_mv(Mv, NegFreeMv),
	convert_to_acyclic_on_mv(Premises, AcPremises, NegFreeMv, R1, R2, R3).

% covers the case where the given list of premises is already acyclic. 
convert_to_acyclic(Premises, Mv, Premises) :-
	\+ is_cyclic_on_mv(Premises, Mv).

%% 	convert_to_acyclic_on_mv(+CPremises, -AcPremises, +Mv, +R1, +R2, +R3)
%
% Converts the cyclic premises to acyclic w.r.t. the metavariable Mv by applying
% weakening in the occurrences of -Mv.
% 
% +CPremiseis	- a list of cyclic premises w.r.t. the metavariable Mv
% -AcPremises 	- the computed list of acyclic premises
% +Mv			- the metavariable causing the cyclicity
% +R1			- the number of premises having multiple occurrences of +Mv
% +R2			- the number of premises having multiple occurrences of -Mv
% +R3			- the number of premises having occurremces of both +Mv and -Mv
%
% eliminates the multiple occurrences of -Mv if there are premises with multiple 
% occurrences of both +Mv and -Mv 
convert_to_acyclic_on_mv(CPremises, AcPremises,  Mv, R1, R2, _):- 
	R1 > 0,
	R2 > 0,
	apply_weakening_on_mv(CPremises, -Mv, AcPremises).

% eliminates the multiple occurrences of -Mv if there are premises with multiple 
% occurrences of +Mv and and also with mixed occurrences of  +Mv and -Mv
convert_to_acyclic_on_mv(CPremises, AcPremises,  Mv, R1, R2, R3):- 
	R1 > 0,
	R2 = 0,
	R3 > 0,
	apply_weakening_on_mv(CPremises, -Mv, AcPremises).

% eliminates the multiple occurrences of Mv if there are premises with multiple 
% occurrences of -Mv and and also with mixed occurrences of  +Mv and -Mv	
convert_to_acyclic_on_mv(CPremises, AcPremises,  Mv, R1, R2, R3):- 
	R1 = 0,
	R2 > 0,
	R3 > 0,
	apply_weakening_on_mv(CPremises, Mv, AcPremises).
	
% eliminates the multiple occurrences of -Mv if there are premises with mixed 
% occurrences of  +Mv and -Mv
convert_to_acyclic_on_mv(CPremises, AcPremises,  Mv, R1, R2, R3):- 
	R1 = 0,
	R2 = 0,
	R3 > 0,
	apply_weakening_on_mv(CPremises, -Mv, AcPremises).

%% apply_weakening_on_mv(+Premises, +Mv, -AcyclicPremises)
%
% Converts the given list of cyclic rules to acyclic by applying 
% the weakening rule to the given metavariable. 
%
% +Premises 		- list of cyclic premises premises w.r.t Mv
% +Mv			- possibly negated metavariable
% -AcPremises	- list of acyclic premises w.r.t Mv
%
% converts 	[[[[], [a, a, 'Y'+1]]], [[[], [a, -a, 'Y'+2]]], [[[], [a, 'Y'+3]]] ]
% to 		[[[[], [a, a, 'Y'+1]]], [[[], [-a, 'Y'+2]]], [[[], [a, 'Y'+3]]] ]
%
apply_weakening_on_mv([], _, []).

apply_weakening_on_mv([[[_, S]]|T], Mv, AcPremises) :-
	apply_weakening_on_mv(T, Mv, AcPremises0),
	neg_free_mv(Mv, NegFreeMv),
	has_mixed_polarities_mv(S, NegFreeMv, 1),
	removeAll(Mv, S, Ac),
	append([[[[], Ac]]], AcPremises0, AcPremises).

% converts 	[[[[], [a, a, 'Y'+1]]], [[[], [-a, -a, 'Y'+2]]], [[[], [a, 'Y'+3]]] ]
% to 		[[[[], [a, a, 'Y'+1]]], [[[], [-a, 'Y'+2]]], [[[], [a, 'Y'+3]]] ]	
apply_weakening_on_mv([[[_, S]]|T], Mv, AcPremises) :-
	apply_weakening_on_mv(T, Mv, AcPremises0),
	count_repetitions(S, Mv, NumRepetitions),
	NumRepetitions > 1,
	removeAll(Mv, S, Ac),
	append([Mv], Ac, Ac1),
	append([[[[], Ac1]]], AcPremises0, AcPremises).
	
apply_weakening_on_mv([[[_, S]]|T], Mv, AcPremises) :-
	apply_weakening_on_mv(T, Mv, AcPremises0),
	count_repetitions(S, Mv, NumRepetitions),
	NumRepetitions = 1,
	neg_free_mv(Mv, NegFreeMv),
	has_mixed_polarities_mv(S, NegFreeMv, 0),
	append([[[[], S]]], AcPremises0, AcPremises).

apply_weakening_on_mv([[[_, S]]|T], Mv, AcPremises) :-
	apply_weakening_on_mv(T, Mv, AcPremises0),
	count_repetitions(S, Mv, NumRepetitions),
	NumRepetitions = 0,
	append([[[[], S]]], AcPremises0, AcPremises).

%% 	count_having_multi_occurrence(+Premises, +Mv, -Counter)
%
% Counts the number of premises having more than one occurrence of the given 
% metavariable. 
%
% +Premises 		- a list of premises
% +Mv			- a metavariable to look for.
% -Counter		- the number of premises having more than one 
%					occurrence of the metavariable. 
count_having_multi_occurrence([], _, 0).

count_having_multi_occurrence([[[_, S]]|T], Mv, NumPrMultiMv) :-
	count_having_multi_occurrence(T, Mv, Num0),
	count_repetitions(S, Mv, NumRepetitions),
	( (NumRepetitions > 1)  -> 
		NumPrMultiMv is Num0 + 1 
		;NumPrMultiMv is Num0
	).

%% count_having_mixed_occurrence(+Premises, +Mv, -NumPremisesMixedMv)
%
% Counts the number of premises having occurrences of both +Mv and -Mv.
% 
% +Premises				- list of premises to be checked
% +Mv					- the metavariable whose occurrences will be chekced
% -NumPremisesMixedMv	- the computed number of premises with mixed occurrence of Mv
count_having_mixed_occurrence([], _, 0).

count_having_mixed_occurrence([[[_, S]]|T], Mv, NumPrMixedMv) :-
	count_having_mixed_occurrence(T, Mv, Num0),
	has_mixed_polarities_mv(S, Mv, R), 
	NumPrMixedMv is Num0 + R.
	
%% has_mixed_polarities_mv(+Succedent, +Mv, -Flag)
%
% Checks if the given succedent has positive and negative occurrence(s) 
% of the given metavariable.
%
% +S 		- a list representing the succedent of a sequent 
% +Mv 		- a negation free metavariable
% -Flag		- a flag evaluated to 0 if the suequent does not contain positive
%			  and negative occurrences of Mv, and 1 otherwise. 
has_mixed_polarities_mv(S, Mv, 0) :-
	\+ member(Mv, S).
	
has_mixed_polarities_mv(S, Mv, 0) :-
	\+ member(-Mv, S).

has_mixed_polarities_mv(S, Mv, 1) :-
	member(Mv, S),
	member(-Mv, S).