%% mall_analytic
% 
% Applies the rule completion procedure i.e. transforms a (hyper)structural rule
% to analytic by cutting the metavariable in all possible combinations. Furthermore, 
% checks the acyclicity condition and transforms the premises into acyclic by applying
% weakening in the metavariables that violate the acyclic property. 

%% htoa_ms_r(+HyperstructualRule, -AnalyticRule, -CyclicCount)
%
% Transforms a (hyper)structural rule into an analytic one. Furthermore, it counts the number  
% of premises that do not satisfy the acyclicity property.
% Expected datastructure for P0 is:
%	 [	
%		[	
%			[[], [p11, p12, ... p1n]], ..., [[], [...]]  
%		],
%		... 
%	]
%expected datastructure for C0 is: [ [[], [c11, c12, ... c1n]], ..., [[], [ck1, ck2, ..., ckm]] ]
%
% +HyperstructuralRule 	- hypersetructural rule to be transformed
% -AnalyticRule			- the generated cyclic rule
% -CyclicCount			- the number of cyclic premises
	
htoa_ms_r([P0, C0], [P1, C1], NumCyclic) :-
	C1 = C0,
	cut(P0, P1, NumCyclic).
	
%% cut(+InitPremises, -MetaVarFreePremises, -NumCyclic)
%
% Cuts the occurrences of the metavariables from the premises.
% +InitPremises 			- premises of the rule resulting from atoh procedure.
% -MetaVarFreePremises	- premises having no metavariables in the conclusion.
% -NumCyclic 			- number of cyclic premises 
cut(P0, P0, 0):-
	are_metafree_premises(P0).
	
cut(P0, P2, C2) :-
	\+ are_metafree_premises(P0), 
	have_metavar(P0, MV),
	neg_free_mv(MV, NegFreeMv), % remove the negation if any from the Mv
	cut_metavariable(P0, P1, NegFreeMv, C1), 
	cut(P1, P2, C0),
	C2 is C0 + C1.
	
%% cut_metavariable(+Premises, -Premises, +Mv)
%
% Cuts the given metavariable from the given list of premises. Furthermore, checks
% whether the given premise is cyclic with respect to the given metavariable. 
% +Premises 			- list of premises before cutting the metavariable
% -Premises			- list of premises after cutting the metavariable 		
% +Mv 				- a metavariable occurring in the premises.
% -CyclicFlag		- an indicator of the acyclicity of the premises w.r.t. MV
cut_metavariable(P0, P2, MV, 0) :- 
	\+ is_cyclic_on_mv(P0, MV),
	collect_by_mv(P0, MV, P1, NumPos, NumNeg), 
	cut_mv_from_premise(P0, P1, P2, NumPos, NumNeg, MV). 
	
cut_metavariable(P0, P2, MV, 1) :- 
	is_cyclic_on_mv(P0, MV),
	convert_to_acyclic(P0, MV, AcP0),
	collect_by_mv(AcP0, MV, P1, NumPos, NumNeg), 
	cut_mv_from_premise(AcP0, P1, P2, NumPos, NumNeg, MV). 
	
%% collect_by_mv(+Premises, +Mv, -PremisesWithMv, -NumPos, -NumNeg)
%
% Collects the list of premises having at least one occurrence of the
% given metavariable. Considers both positive and negative occurrences.
% Furthermore, counts the number of cyclic premises.  
% 
% +Premises 			- the list of all premises to search in
% +Mv				- the metavariable
% -PremisesWithMv	- the list of premises having an occurrence of Mv
% -NumPos			- number of premises having occurrence(s) of positive Mv
% -NumNeg 			- number of premises having occurrence(s) of negative Mv
collect_by_mv([], _, [], 0, 0).

% covers the case where a premise has a positive occurrence of Mv
collect_by_mv([H|T], Mv, P1, A1, S0) :-
	contains_mv(H, Mv), 
	\+ contains_mv(H, -Mv),
	collect_by_mv(T, Mv, P0, A0, S0),
	append([H], P0, P1),
	A1 is A0 + 1.

% covers the case where a premise has a negative occurrence of Mv
collect_by_mv([H|T], Mv, P1, A0, S1) :-
	contains_mv(H, -Mv),
	\+ contains_mv(H, Mv), 
	collect_by_mv(T, Mv, P0, A0, S0),
	append([H], P0, P1),
	S1 is S0 + 1.

% covers the case where a premise has no occurrence of Mv
collect_by_mv([H|T], Mv, P1, A0, S0) :-
	\+ contains_mv(H, Mv), 
	\+ contains_mv(H, -Mv), 
	collect_by_mv(T, Mv, P0, A0, S0),
	append([], P0, P1).

%% contains_mv(+HyperSequent, +Mv)
% 
% A helper predicate for checking whether a hypersequent contains a meta variable.
% +HyperSequent 		- hypersequent to search in
% +Mv 				- a formula metavariable to search for.
contains_mv([[_, S]|_], Mv) :-
	member(Mv, S).
	
contains_mv([[_, S]|T], Mv) :-
	\+ member(Mv, S),
	contains_mv([T], Mv).

%% cut_mv_from_premise(+Premises, +PremisesWithMv, -Premises, +NumPos, +NumNeg, +Mv)
% 
% Cuts the given metavariable in all possible combinations. 
% 
% +Premises 			- the list of original premises.
% +PremisesWithMv	- the list of premises having at least one occurrence of the metavariable
% -Premises			- the new list of premises obtained by cutting the metavariable
% +NunPosMv			- the number of premises containing a positive occurrence of the metavariable
% +NumNegMv			- the number of premises containing a negative occurrence of the metavariable
% +Mv				- the metavariable to be cut
%
cut_mv_from_premise(P0, PrWithMv, P3, NumPosMv, NumNegMv, Mv) :- 
	remove_premises(PrWithMv, P0, P1),		
	separate_premises(PrWithMv, PrPosMv, PrNegMv, Mv),
	( (NumPosMv = 1 ; NumNegMv = 1 ) -> 
		(
			(NumPosMv = 1) -> (
				simple_merge_neg(PrPosMv, PrNegMv, P2, Mv)
				);( 
				simple_merge_pos(PrPosMv, PrNegMv, P2, Mv)
				)
		);( 
			% count number of premises having more than one occurrence of positive metavariable 
			count_premises_with_multi_pos_mv(PrPosMv, Mv, NumMultiPosMv),
			(
				(NumMultiPosMv > 0) -> (
					combinatorial_merge_pos(PrPosMv, PrNegMv, P2, NumNegMv, Mv)
					);(
					combinatorial_merge_neg(PrNegMv, PrPosMv, P2, NumPosMv, Mv)
				)
			)
		)
	),
	append(P1, P2, P3).
	
% simple_merge_pos(+PrPosMv, +PrNegMv, -Premises, +MetaVariable)
simple_merge_pos([], [], [], _).

% when there are no premisises with positive metaVariable
simple_merge_pos([], _, [], _).

% when there are no premisises with negative metaVariable
simple_merge_pos(_, [], [], _).

simple_merge_pos([[[Apos, Spos]]|Tpos], [[[Aneg, Sneg]]|Tneg], P3, Mv) :-
	simple_merge_pos([[[Apos, Spos]]|Tpos], Tneg, P1, Mv),
	simple_merge_pos(Tpos, [[[Aneg, Sneg]]|Tneg], P2, Mv),
	get_repleacing_elements(Mv, Sneg, E),
	replace_elements(Spos, Mv, E, Spos1),
	append( [[[Aneg, Spos1]]], P1, P11),
	append(P2, P11, P3).
	
% simple_merge_neg(+PrPosMv, +PrNegMv, -Premises, +MetaVariable)
simple_merge_neg([], [], [], _).

% when there are no premisises with positive metaVariable
simple_merge_neg([], _, [], _).

% when there are no premisises with negative metaVariable
simple_merge_neg(_, [], [], _).

simple_merge_neg([[[Apos, Spos]]|Tpos], [[[Aneg, Sneg]]|Tneg], P3, Mv) :-
	simple_merge_neg([[[Apos, Spos]]|Tpos], Tneg, P1, Mv),
	simple_merge_neg(Tpos, [[[Aneg, Sneg]]|Tneg], P2, Mv),
	get_repleacing_elements(Mv, Spos, E),
	replace_elements(Sneg, Mv, E, Sneg1),
	append( [[[Aneg, Sneg1]]], P1, P11),
	append(P2, P11, P3).
	
get_repleacing_elements(Mv, S, E) :- 
	remove(Mv, S, E0), 
	remove(-Mv, E0, E).

%% combinatorial_merge_pos(+PremisesPosMv, +PremisesNegMv, -Premises, +NumPremisesNegMv, +MetaVariable)
%
% Covers the case when each premise contains at most one negative occurrence
% of the metavariable. (Positive metavariables may occur more than once
% in some premises). The resulting premises are obtained by cutting the
% given metavariable in all possible combinations.
%
% + PremisesPosMv 	- list of premises with positive occurence(s) of metavariable
% + PremisesNegMv	- list of premises with ONE negative occurece of metavariable
% - Premises			- list of resulting premises (after performing combinatorial merge)
% + NumPremisesNegMv	- number of premises having a negative occurence of metavariable
% + Metavariable		- the given metavariable
%
combinatorial_merge_pos([], [], [], _, _).

combinatorial_merge_pos([], _, [], _, _).

combinatorial_merge_pos(_, [], [], _, _).

combinatorial_merge_pos([[[Apos, Spos]]|Tpos], [Hneg | Tneg], P3, 1, Mv) :-
	combinatorial_merge_pos([[[Apos, Spos]]|Tpos], Tneg, P1, 1, Mv),
	combinatorial_merge_pos(Tpos, Hneg, P2, 1, Mv),
	get_repleacing_elements(Mv, Hneg, E),
	replace_elements(Spos, Mv, E, Spos1),
	append( [[[Apos, Spos1]]], P1, P11),
	append(P2, P11, P3).

combinatorial_merge_pos([[[Apos, Spos]]|Tpos], Pneg0, P2, NumNegPremises, Mv) :-
	combinatorial_merge_pos(Tpos, Pneg0, P1, NumNegPremises, Mv),
	count_repetitions(Spos, Mv, NumRepetitions),
	calc_num_new_premises(NumNegPremises, NumRepetitions, NumNewPremises), 
	flatten(Pneg0, Pneg1), 
	removeAll(-Mv, Pneg1, Pneg2),
	generate_combination(NumRepetitions, Pneg2, 0, NumNewPremises, [], Result),
	create_new_premises([Apos, Spos], Result, NewPremise, Mv),
	append(P1, NewPremise, P2).
	
%% combinatorial_merge_neg(+PremisesNegMv, +PremisesPosMv, -Premises, +NumPremisesPosMv, +MetaVariable)
%
% Covers the case where each premise contains at most one positive occurrence
% of the metavariable. (Negative metavariables may occur more than once
% in some premise). The resulting premises are obtained by cutting the
% given metavariable in all possible combinations.
%
% + PremisesNegMv 		- list of premises with positive occurence(s) of metavariable
% + PremisesPosMv		- list of premises with ONE negative occurece of metavariable
% - Premises			- list of resulting premises (after performing combinatorial merge)
% + NumPremisesPosMv	- number of premises having a negative occurence of metavariable
% + Metavariable		- the given metavariable

combinatorial_merge_neg([], [], [], _, _).

combinatorial_merge_neg([], _, [], _, _).

combinatorial_merge_neg(_, [], [], _, _).

combinatorial_merge_neg([[[Aneg, Sneg]|Tneg]], [Hpos | Tpos], P3, 1, Mv) :-
	combinatorial_merge_neg([[[Aneg, Sneg]]|Tpos], Tpos, P1, 1, Mv),
	combinatorial_merge_neg(Tneg, Hpos, P2, 1, Mv),
	get_repleacing_elements(Mv, Sneg, E),
	replace_elements(Sneg, Mv, E, Sneg1),
	append([ [[Aneg, Sneg1]]], P1, P11),
	append(P2, P11, P3).
	
	
combinatorial_merge_neg([[[Aneg, Sneg]]|Tpos], Ppos0, P2, NumPosPremises, Mv) :-
	combinatorial_merge_neg(Tpos, Ppos0, P1, NumPosPremises, Mv), %todo: change to _neg and test
	count_repetitions(Sneg, -Mv, NumRepetitions),
	calc_num_new_premises(NumPosPremises, NumRepetitions, NumNewPremises), 
	flatten(Ppos0, Ppos1),
	removeAll(Mv, Ppos1, Ppos2),
	generate_combination(NumRepetitions, Ppos2, 0, NumNewPremises, [], Result),
	create_new_premises([Aneg, Sneg], Result, NewPremise, Mv),
	append(P1, NewPremise, P2).
	
%% generate_combination(+NumRepetitions, +PremisesWithMv, +Count, +MaxCount, +ExistingRes, -Res)
%
% NumRepetitions - number of repetitions of the metavariable in the current premise
% PremisesWithMv - premises having an occurrence of negated metavariable
% Count 		 	 - current index of generated combination
% MaxCount		 - maximum number of combinations
% ExistingRes 	 - combinations generated so far
% Res			 - final result.
generate_combination(_, _, NumPosPremises, NumPosPremises, Result, Result).

generate_combination(NumRepetitions, L, 0, NumNewPremises, [], Result):-
	combine_with_repetitions(NumRepetitions, L, X0), 
	generate_combination(NumRepetitions, L, 1, NumNewPremises, [X0], Result).
	
generate_combination(NumRepetitions, L, NumPremises, NumNewPremises, ExistingRes, Res) :-
	NumPremises < NumNewPremises,
	combine_with_repetitions(NumRepetitions, L, X0),
	\+ member(X0, ExistingRes),
	append([X0], ExistingRes, ExistingRes1),
	NumPremises1 is NumPremises+1 ,
	generate_combination(NumRepetitions, L, NumPremises1, NumNewPremises, ExistingRes1, Res).

generate_combination(NumRepetitions, L, NumPremises, NumNewPremises, ExistingRes, Res) :-
	NumPremises < NumNewPremises,
	combine_with_repetitions(NumRepetitions, L, X0),
	member(X0, ExistingRes),
	generate_combination(NumRepetitions, L, NumPremises, NumNewPremises, ExistingRes, Res).	

%% combine_with_repetitions(+N, +Premises, -Result) 
%
% +N 		- number of repetitions of metavariable
% +L 		- list of premises
% -R		- combined result
%
combine_with_repetitions(0,_,[]).
combine_with_repetitions(N,[X|T],[X|RComb]) :-
	N>0,
	N1 is N-1,
	combine_with_repetitions(N1,[X|T],RComb).
	
combine_with_repetitions(N,[_|T],RComb) :-
	N>0,
	combine_with_repetitions(N,T,RComb).

%% create_new_premises(+Sequent, +Premises, -PNew, -Mv) 
%
% Removes the occurrences of the Mv from the given sequent  and merges it with
% each of the premises in the given list. 
% 
% +Sequent 	- a sequent with occurrences of Mv
% +Premises	- a list of newly generated premises
% -PNew		- the premises obtained from the merging step
% +Mv		- the metavariable to be removed
%
create_new_premises(_,[],[],_).
create_new_premises([Ante,Succ], [H|T], PNew, MV) :-
	create_new_premises([Ante,Succ], T, PNew1, MV),
	removeAll(MV,Succ,Succ01),
	removeAll(-MV,Succ01,Succ1),
	append(H,Succ1,Succ2),
	append([[[Ante,Succ2]]],PNew1,PNew).

%% calc_num_new_premises(+NumPremisesWithMv, +NumRepeptitions, -NumNew) 
%
% Calculates the number of the premises obtained by cutting a premise with 
% multiple occurrences of positive (negative) metavariable with the premises having one 
% occurrence of negative (positive) metavariable. 
%
% +NumPos(NumNeg)	- number of premises having one occurrence of positive(negative) metavariable
% +NumRepetitions 	- number of repetitions of negative(positive) metavariable in the premise
% -NumNew			- number of premises obtained by cutting the metavariable in all possible combinations. 
%
calc_num_new_premises(NumPos, NumRepetitions, NumNew) :-
	fac(NumPos+NumRepetitions-1,R1),
  	fac(NumPos-1,R2),
  	fac(NumRepetitions,R3),
  	NumNew is R1/(R2*R3).

%% fac(+N, -F)
% 
% Calculates the factorial of the given number. 
% 
% +N 	- an integer whose factorial shall be computed
% -F		- the factorial of the given integer.
%
fac(0,1).
fac(N,F) :-
	N1 is N-1,
	fac(N1,F1),
	F is N*F1.
	
%% separate_premises(+Premiseses, -PremisesPosMv, -PremisesNegMv, +Mv)
%
% Separates a list of premises based on the occurrences of the given metavariable. 
% Note that the provided list of premises should be acyclic. 
%
% +Premises 		- the list of premises to be split
% -PremisesPosMv	- the list of premises having an occurrence of +Mv
% -PremisesNegMv	- the list of premises having an occurrence of -Mv
% +Mv 			- the metavariable used to separate premises
%
separate_premises([], [], [], _).

separate_premises([H|T], PrPosMv1, PrNegMv0, Mv) :-
	\+ contains_mv(H, -Mv), 
	contains_mv(H, Mv), 
	separate_premises(T, PrPosMv0, PrNegMv0, Mv),
	append([H], PrPosMv0, PrPosMv1).
	
separate_premises([H|T], PrPosMv0, PrNegMv1, Mv) :-
	contains_mv(H, -Mv),
	separate_premises(T, PrPosMv0, PrNegMv0, Mv),
	append([H], PrNegMv0, PrNegMv1).
	
%% remove_premises(+ToBeRemoved, +Premises, -Remaining)
%
% Find the list of the remaining premises after removing the elements of the first
% list from the second one.
% 
% +ToBeRemoved 		- the premises to be removed
% +Premises			- the full list of premises potentially containing elements from ToBeRemoved
% -Remaining			- the remaining elements of Premises after removing ToBeRemoved
remove_premises([], Premises, Premises).

remove_premises([H|T], PrWithMv, P2) :-
	remove(H, PrWithMv, P1),
	remove_premises(T, P1, P2).

%% have_metavar(+Premises, +Mv)	
% 
% Checks the premises for occurrences of the metavariable.
%
% +Premises 		- the list of premises to be checcked
% +Mv			- the metavariable being checked
have_metavar([H|_], Mv) :-
	contains_metavariable(H, Mv).

have_metavar([H|T], Mv) :-
	\+ contains_metavariable(H, Mv),
	have_metavar(T, Mv).
	
%% contains_metavariable(+Premise, +Mv)
%
% Checks whether the given premise has an occurrence of the given metavariable
%
% +Premise 		- a (hyper)sequent premise
% +Mv			- the metavariable to be checked
contains_metavariable([ [[], [H| _]] |_], H) :-
	H \= 'Y'+ _, 
	H \= -('Y'+_).
	
contains_metavariable([ [[], ['Y'+_ | T1]] |T], MV) :-
	contains_metavariable([ [[], T1] |T], MV).
	
contains_metavariable([ [[], [H| T1]] |T2], MV) :-
	H \= 'Y'+_,
	H \= - ('Y'+_),
	\+ atomic(H),
	compound(H),
	contains_metavariable([[[], T1]|T2], MV).
	
contains_metavariable([ [[], []] |T], MV) :-
	contains_metavariable(T, MV).

%% are_metafree_premises(+Premises).
%
% checks whether any of the premises contains any metavariable.
% +Premises - 	list of premises to be checked.
are_metafree_premises([]).

are_metafree_premises([H|T]) :- 
	are_metafree_premises(T),
	is_metafree_premises(H).
	
%% is_metafree_premises(+Premise).
%
% checks whether the given premise contains any metavariable.
% +Premise - 	premise to be checked.
is_metafree_premises([]).

is_metafree_premises([[_, S]|T]) :-
	is_metafree_succedent(S),
	is_metafree_premises(T).
	
is_metafree_antecedent([]).

%in this extension, the antecedent is always en empty list
%is_metafree_antecedent(['Y'+_ | T]) :-
%	is_metafree_antecedent(T).
%is_metafree_antecedent(['Y'+_ | T]) :-
%	is_metafree_antecedent(T).

is_metafree_succedent([]).
%is_metafree_succedent(['Y'+_ | T]) :-
%	is_metafree_succedent(T).
	
is_metafree_succedent([H|T]) :-
	H \= 'Y'+_,
	H \= - ('Y'+_ ),
	\+ atomic(H),
	compound(H),
	neg_free_mv(H, H2),
	\+ is_atom(H2),
	is_metafree_succedent(T).

is_metafree_succedent([]).
	
is_metafree_succedent(['Y'+_ | T]) :-
	is_metafree_succedent(T).
	
is_metafree_succedent([-( 'Y'+_ ) | T]) :-
	is_metafree_succedent(T).

%% neg_free_mv(+Mv, -NegFreeMv)
%
% Removes  (if any) the negative polarity from the metavariable.
%
% +Mv 			- a potentially negated metavariable 
% -NegFreeMv	- the same metavariable free from any negative polarity
%
neg_free_mv(MV, NegFreeMv):-
	(MV = -X -> 
		NegFreeMv = X
		;
		NegFreeMv = MV
		).
	
%% count_premises_with_multi_pos_mv(+Premises, +Mv, -NumPrMultiMv)
%
% Count the premises having multiple occurrences of the positive metavariable Mv
%
% +Premises 		- the list of premises to be checked
% +Mv			- the metavariable to look for
% -NumPrMultiMv	- the number of the premises containing multiple (strictly more than one)
%				occurrences of the metavariable Mv
count_premises_with_multi_pos_mv([], _,0).

count_premises_with_multi_pos_mv([[[_, S]]|T], Mv, NumPrMultiMv) :-
	count_premises_with_multi_pos_mv(T, Mv, Num0),
	count_repetitions(S, Mv, NumRepetitions),
	( (NumRepetitions > 1)  -> 
		NumPrMultiMv is Num0 + 1 
		;NumPrMultiMv is Num0
	).
	
		
%% replacElements(+List, +ReplaceElement, +NewElement(s), -List)
%%		replaces an element with a new element (which is a list)
replace_elements([], _, _, []).

replace_elements([R|T], R, E, NL1) :-
	replace_elements(T, R, E, NL0),
	append(E, NL0, NL1).
	
replace_elements([-R|T], R, E, NL1) :-
	replace_elements(T, R, E, NL0),
	append(E, NL0, NL1).
	
replace_elements([H|T], R, E, NL1) :-
	H \= R,
	H \= -R, 
	replace_elements(T, R, E, NL0),
	append([H], NL0, NL1).
	
negate_elements([], []).
negate_elements([H|T], P2) :-
	negate_elements(T, P1),
	append([-H], P1, P2).
		