%% mall_axiom2rule
%
% Puts together the transformation steps implemented on each module. 
% Furthermore, transforms (if needed) the input axiom into a set of
% equivalent axioms. The result of the transformation is printed in the 
% console and also in LaTeX format. 
%


%% axiom_to_rule(+Axiom, +Class)
%
% converts an axiom of class up to p_prime(3) to a list of analytic rules.
% 
% + Axiom 		- the axiom to be transformed.
% + Class		- the class of the axiom to be transformed.
%
% covers the axioms of the class P1, N1 and N2
axiom_to_rule(Axiom, Class) :-
	member(Class, [p(1), n(1), n(2)]),
	to_n2_normal(Axiom, N2normal),
	to_rule(N2normal, Class, Rules, NARules, NumCyclic), 
	tex_out(Axiom, Rules, NARules, Class, NumCyclic).

% covers the axioms of the class P2 and P'3
axiom_to_rule(Axiom, Class) :-
	member(Class, [p(2), p_prime(3)]),
	to_hypersequent(Axiom, Hyperseq),
	reverseList(Hyperseq, [], Hyperseq1),
	to_hypersequent_list([Hyperseq1], HyperseqList),
	reverseList(HyperseqList, [], HyperseqList1),
	to_rule(HyperseqList1, Class, AnalyticRule, NARule, NumCyclic),
	tex_out(Axiom, AnalyticRule, NARule, Class, NumCyclic).

% covers the case of unsupported classes	
axiom_to_rule(_, Class) :-
	\+ member(Class, [p(1), p(2), p_prime(3), n(1), n(2)]),
	writef('\n The algorithm does not work for this class yet.'),
	tex_out(_, _, _, Class, 0).

%% to_rule(+ListOfNormalAxioms, +Class, -ListOfRules)
%
% converts each axiom on the given list of normal axioms to an analytic rule.
%
% +ListOfNormalAxioms 	- the list of axioms to be transformed
% +Class 				- class of the axioms
% -ListOfRules 			- the list of the generated rules
to_rule([], _, [], [], 0).
to_rule([H|T], Class, Analytic, NAnalytic, NumCyclic):- 
		aths_ms_r(H, Class, Hyperstructural),
		htoa_ms_r(Hyperstructural, Analytic1, NumCyclic1),
		writef('\n\nEquivalent Analytic Rule: \n'),
		print_rule(Analytic1, Class),
		to_rule(T, Class, Analytic2, NAnalytic2, NumCyclic2),
		append([Analytic1], Analytic2, Analytic),
		append([Hyperstructural], NAnalytic2, NAnalytic),
		NumCyclic is NumCyclic2 + NumCyclic1.
		
%% to_n2_normal(+Axiom, -ListAxioms)
%
% Converts an axiom of class N2 to a list of N2-normal axiom which is equivalent to.
% 
% +Axiom 		- an axiom in N2
% -ListAxioms	- a list of N2-normal axioms which are equivalent with the given axiom.
to_n2_normal(Ax1 & Ax2 , N2normal1) :-
	to_n2_normal(Ax1, N2normal0), 			% & concatenations in a N2 axiom  
	append([Ax2], N2normal0, N2normal1).	% have to be splited into a list of N2-normal axioms

to_n2_normal(Ax, N2normal):-
	atomic(Ax),
	append([Ax], [], N2normal).
	
to_n2_normal(Ax, N2normal):-
	compound(Ax),
	append([Ax], [], N2normal).

%% to_hypersequent(+Axiom, -Hypersequent)
%
% converts the given list of v concatenated axioms into a hypersequent of 
% N2 normal axioms.
to_hypersequent(Ax1 v Ax2, Hyperseq1) :-
	to_hypersequent(Ax1, Hyperseq0),
	Ax2 = Ax21 & 1 ,
	append([[[],[Ax21]]], Hyperseq0, Hyperseq1).
	
to_hypersequent(Ax1 v Ax2, Hyperseq1) :-
	to_hypersequent(Ax1, Hyperseq0),
	Ax2 \= _ & 1 ,
	append([[[],[Ax2]]], Hyperseq0, Hyperseq1).
	
to_hypersequent(Ax, Hyperseq) :-
	atomic(Ax),
	append([[[],[Ax]]], [], Hyperseq).

to_hypersequent(Ax, Hyperseq) :-
	compound(Ax),
	(Ax = Ax01 & 1) -> 
	( append([[[],[Ax01]]], [], Hyperseq)
	);( append([[[],[Ax]]], [], Hyperseq)).
	
%% to_hypersequent_list(+Hypersequent, -Hypersequent)
%
% checks whether the hypersequent contains applications of * and, if yes,
% creates more hypersequents
to_hypersequent_list([],[]). 
to_hypersequent_list([H|T], New) :-
	member(([[[],[F1 * F2]]]), H),
	remove(([[[],[F1 * F2]]]), H, H1),
	append([[[],[F1]]], H1, H2),
	append([[[],[F2]]], H1, H3),
	append([H3], [H2], H4),
	to_hypersequent_list(H4, New1),
	to_hypersequent_list(T, New2),
	( New2 = [] ->
	  	New = New1
		; append(New2,New1,New)
	).

to_hypersequent_list([H|T], New) :-
	\+ member(([[],[_ * _]]), H),
	to_hypersequent_list(T, New1),
	append([H],New1,New).
