%% mall_hierarchy
%
% Computes the class in the substructural hierarchy where the input axiom falls into.
%


% P_0 = N_0 --> Atom
% P_n --> 	N_n-1 	|  P_n v P_n	|  P_n * P_n 	| 1 	| 0
% N_n -->	P_n-1 	|  N_n + N_n	|  N_n & N_n 	| bot 	| top

% P_3^' --> N_2 & 1 | P_3^' v P_3^' | P_3^1 * P_3^' | 1 	| 0


%% is_in_class(+Axiom, -Class)
% 
% finds the least class of the given axiom. 
is_in_class(Axiom, C) :-
	is_pos_class(Axiom, 0, P),
	is_neg_class(Axiom, 0, N),
	( P < N ->  (
		( P = 3,  is_pos_3prime_class(Axiom)) -> C = p_prime(P); C = p(P)
	);( C = n(N)) ).

%% is_pos_class(+Axiom, +StartFrom, -LeastPosClass)
% 
% finds the least positive class where the axiom falls into. 
%
is_pos_class(A, P0, P0) :- 
	is_pos_axiom(A, P0).
	
	
is_pos_class(A, P0, P2) :-
	\+ is_pos_axiom(A, P0), 
	P1 is P0 + 1, 
	is_pos_class(A, P1, P2).
	
%% is_neg_class(+Axiom, +StartFrom, -LeastPosClass)
% 
% finds the least negative class where the axiom falls into. 
%
is_neg_class(A, N0, N0) :-
	is_neg_axiom(A, N0).

is_neg_class(A, N0, N2) :-
	\+ is_neg_axiom(A, N0), 
	N1 is N0 + 1, 
	is_neg_class(A, N1, N2).
	
%% is_pos_3prime_class(+Axiom)
% 
% checks wether the given axiom falls into class P3'. 
%
is_pos_3prime_class(A) :-
	is_pos_3prime_axiom(A).

%% is_pos_axiom(+Axiom, +Class)
%
% checks wether the given axiom falls into the given positive class.
%
is_pos_axiom(A, 0) :-     	% P_0 :- A
	is_atom(A).				% covers the case when the axiom is an atom
	
is_pos_axiom(-A, 0) :-     	% P_0 :- -A
	is_atom(A).				% covers the case when the axiom is an -atom
	
is_pos_axiom(1, N) :- 	    % P_{n+1} :- 1
	N > 0.					% covers the case when the axiom is the constant 1
	
is_pos_axiom(0, N) :- 	    % P_{n+1} :- 0
	N > 0.					% covers the case when the axiom is the constant 0

%% 
% covers the case when the axiom is in negative class n-1
%
is_pos_axiom(A, N1) :-    	% P_{n+1} :- N_n
	N1 > 0,
    N is N1-1,
    is_neg_axiom(A, N).	

%%   
% covers the case when the axiom is a 'v' concatenation 
% of two axioms in class Pn
%
is_pos_axiom(A v B, N) :- 	% P_{n+1} :- P_{n+1} v P_{n+1}
	N > 0,
    is_pos_axiom(A, N),
    is_pos_axiom(B, N).		
   
%%   
% covers the case when the axiom is a '*' concatenation of two axioms in class Pn
% 
is_pos_axiom(A * B, N) :- 	% P_{n+1} :- P_{n+1} * P_{n+1}
	N > 0,
	is_pos_axiom(A, N),
	is_pos_axiom(B, N).

%%
% 1 and 0 are P3 prime

is_pos_3prime_axiom(1).
is_pos_3prime_axiom(0).

%%
% N2 & 1 is P3 prime
is_pos_3prime_axiom(A & 1) :-
	is_neg_axiom(A, 2).

%%
% P3_prime v P3_prime is also P3_prime
is_pos_3prime_axiom(A v B) :-
	is_pos_3prime_axiom(A),
	is_pos_3prime_axiom(B).

%%
% P3_prime * P3_prime is also P3_prime
is_pos_3prime_axiom(A * B) :-
	is_pos_3prime_axiom(A),
	is_pos_3prime_axiom(B).

%% is_neg_axiom(+Axiom, +Class)
%
% checks wether the given axiom falls into the given negative class.
%	
is_neg_axiom(A, 0) :- 
	is_atom(A).
	
is_neg_axiom(-A, 0) :- 
	is_atom(A).

is_neg_axiom(bot, N) :-
	N > 0.
	
	
is_neg_axiom(top, N) :-
	N > 0.
	
is_neg_axiom(A, N1) :-
	N1 > 0, 
	N is N1 - 1, 
	is_pos_axiom(A, N).

is_neg_axiom(A & B, N) :- 
	N > 0,
	is_neg_axiom(A, N), 
	is_neg_axiom(B, N).
	
	
is_neg_axiom(A + B, N) :- 
	N > 0,
	is_neg_axiom(A, N), 
	is_neg_axiom(B, N).