%% mall_hyperstructural
%
% Transforms an axiom to a list of (hyper)structural rules. 
%
% The used data structure:
% A sequent is represented as a list of two elements: [Antecedent, Succedent]
% 	- The antecedent is usually an empty list.
% 	- The succedent is a flat list 
%	E.g. a sequent could be: [[], [a, b, c, ...]]
% A hypersequent is represented as a list of sequents.
% A rule is a list of two elements: [Premises, Conclusion]
% 	- Premises is a list of hypersequents
%	- Conclusion is a single hypersequent


%% convention
% 1		 -->  1
% 0 	 	 -->  0
% \oplus -->  v
% \otimes --> *
% &		 -->  &
% par  	 -->  +

%% aths_ms_r(+Axiom, +Class, -SequentRule)
% Transforms an axiom to a (hyper)sequent rule. Applies the procedure given 
% in the proof of Theorem 4.2 in [CST09]. Covers axioms up to level N2 of the 
% substructural hierarchy.
% 
% + Axiom 		- the axiom to be transformed.
% + C 			- the class where the axiom belongs to.
% - Structural 	- the resulting structural rule in sequent calculus. 
aths_ms_r(Axiom, C, Structural) :- 
	member(C, [p(1), n(1), n(2)]),
	apply_to_conclusion([[[],[Axiom]]], C1),
	% C1 has the structure: [[], [Ax1, Ax2, ..., Axn]]
	apply_equivalence_lemma([], P2, C1, C2, _),
	apply_invertible_to_premises(P2, P3),
	Structural = [P3, C2].
	
% covers axioms up to level P'3 of the substructural hiearachy Theorem 5.3 in [CST09].
aths_ms_r(HS, C, Structural) :- 
	member(C, [p(2), p_prime(3)]),
	apply_to_conclusion(HS, C1), 
	apply_equivalence_lemma([], P2, C1, C2, _),
	apply_invertible_to_premises(P2, P3),
	Structural = [P3, C2].

%% apply_to_conclusion(+Conclusion, -Conclusion)
%
% Applies the invertible rules to a hypersequent representing the conclusion
% of a rule. The invertible rules are &, +, \bot, \top. 
%
% +/- Conclusion 	- conclusion before/after appling the invertible rules
apply_to_conclusion([], []).

apply_to_conclusion([H|T], C3) :-
	apply_to_conclusion(T, C1),
	apply_invertible_to_conclusion_secuence(H, C2),
	append([C2], C1, C3).
	
%% apply_invertible_to_conclusion_secuence(+SequentConclusion, -SequentConclusion)
%
% Applies the invertible rules to a sequent of the conclusion. 
%
% +/- SequentConclusion 	- a sequent of the conclusion before/after applying the invertible rules.
%
% in this case: |- X, a + b   =>  |- X, a , b
apply_invertible_to_conclusion_secuence([A, S0], C3) :-
	member(Ax1 + Ax2 , S0),
	remove(Ax1 + Ax2, S0, S1),
	append([Ax1 , Ax2], S1, S2),
	apply_invertible_to_conclusion_secuence([A, S2], C3).

% in this case: |- X, bot   =>  |- X
apply_invertible_to_conclusion_secuence([A, S0], C2) :- 
	member(bot, S0), 
	remove(bot, S0, S1), 
	apply_invertible_to_conclusion_secuence([A, S1], C2).
	
% in this case: |- X, top   =>  |-	
apply_invertible_to_conclusion_secuence([A, S0], C1) :- 
	member(top, S0), 
	C1 = [A, []].

apply_invertible_to_conclusion_secuence([A, [Axiom|T]], [A, [Axiom|T]]) :-
	atomic(Axiom).
	
apply_invertible_to_conclusion_secuence([A, [Axiom|T]], [A, [Axiom|T]]) :-
	compound(Axiom).


%% apply_equivalence_lemma(+Premises, -Premises, +Conclusion, -Conclusion, -Index)
%
% Applies the equivalence Lemma to each sequent in the conclusion. See Lemma 5.1 in [CST09],
% which is a generalization to hypersequent calculus of Lemma 4.1. Terminates when every 
% formula in the succedent of each sequent of the conclusion is replaced with a fresh 
% multiset variable and a new premise is introduced for it. 
% 
% +Premises 		- List of premises before utilizing the equivlence Lemma
% -Premises 		- List of premises after utilizing the equivalence Lemma
% +Conclusion 	- The conclusion before utiliziing the equivalence Lemma
% -Conclusion 	- The conclusion after utilizing the equivalence Lemma.
% -Index			- A counting index
% 
% Expected datastructure for the conclusion is: [ [[],[....]], ...., [[],[....]]]
%
apply_equivalence_lemma(_, [], [], [], 1).

apply_equivalence_lemma(P0, P3, [H|T], C3, N2) :-
	%H has the structure: [[], [Ax1, Ax2, ..., Axn]]
	apply_equivalence_lemma(_,  P1, T, C1, N0), 
	apply_to_mc_sequent(P0, P2, H, C2, N0, N1),
	N2 is N0 + N1,
	append([C2], C1, C3),
	append(P2, P1, P3).

%% apply_to_mc_sequent(+Premises, -Premises, +Sequent, -Sequent, +Index, -Index)
%
% Applies the equivalence Lemma to a single sequent of the conclusion. See Lemma 4.1 in [CST09].
%
% +/- Premises 		- list of premises before/after applying equivalence Lemma
% +/- Conclusion 	- the conclusion before/after applying the equivalence Lemma 
% +/- Index 			- counting index offset for ensuring uniqueness of the generated formula variables
%
% Expected datastructure for C0 is: [A,[...]] , where A is usually an empty list
apply_to_mc_sequent(P0, P2, [A, S0], C1, N0, N1) :-
	set_as_premise(S0, S1, P1, N0, N1),
	C1 = [A, S1],
	append(P1, P0, P2).

%% set_as_premise(+ConclusionSequent, -ConclusionSequent, -NewPremises, -Index)
%
% Replaces the formula variables of a sequent in the conclusion with multiset variables
% and introduces a new premise for each of them.
%
% +/- ConclusionSequent 	- a sequent appearing in the conclusion before/after replacing the formula variables.
% - NewPremises 			- a list of the new premises.
% - Index 				- a counter used as a suffix for the fresh multiset variables. 
set_as_premise([], [], [], _, 0).

set_as_premise([H|T], C1, P1, N0, NA1) :-
	N1 is N0 + 1, 
	set_as_premise(T, C0, P0, N1, NA0),
	NA1 is NA0 + 1,
	append(['Y'+N0], C0, C1),
	%De Morgan laws can be applied in this point, because the added premise always starts with negation.
	apply_de_morgan(-H, H2), 
	append([[[[],['Y'+N0, H2]]]], P0, P1).

%% apply_de_morgan(-Axiom, +Axiom)
%
% Applies de Morgan law to the given axiom.
%
% +/- Axiom - an axiom before and after applying de Morgan laws.
apply_de_morgan(-( Ax1 v Ax2 ), Axiom) :- 
	apply_de_morgan(-Ax1, Ax3),
	apply_de_morgan(-Ax2, Ax4),
	Axiom = Ax3 & Ax4 .
	
apply_de_morgan(-( Ax1 & Ax2 ), Axiom) :- 
	apply_de_morgan(-Ax1, Ax3),
	apply_de_morgan(-Ax2, Ax4),
	Axiom = Ax3 v Ax4 .
	
apply_de_morgan(- ( Ax1 * Ax2 ), Axiom) :-
	apply_de_morgan(-Ax1, Ax3),
	apply_de_morgan(-Ax2, Ax4),
	Axiom = Ax3 + Ax4 .
	
apply_de_morgan(- ( Ax1 + Ax2 ), Axiom) :-
	apply_de_morgan(-Ax1, Ax3),
	apply_de_morgan(-Ax2, Ax4),
	Axiom = Ax3 * Ax4 .
	
apply_de_morgan( -(-Ax1) , Axiom) :-
	Axiom = Ax1 .

apply_de_morgan(-( Ax1 ), Ax2) :-
	is_atom(Ax1),
	Ax2 = - (Ax1).
	
apply_de_morgan(-( 1 ), Ax2) :-
	Ax2 = 0.
	
apply_de_morgan(-( 0 ), Ax2) :-
	Ax2 = 1.
	
apply_de_morgan(-( bot ), Ax2) :-
	Ax2 = top.
	
apply_de_morgan(-( top ), Ax2) :-
	Ax2 = bot.

apply_de_morgan(Ax, Ax) :-
	\+ Ax = -(_).

%% apply_invertible_to_premises(+Premise, -Premise)
%
% Applies invertible rules (+ and &) to the given list of premises. Terminates 
% when there are no more logical connectives occurring in the premises.
%
% +/- Premise - list of premises before/after applying the invertible rules.
%
% the expected datastructure for the list of premises is: 
% [ [[A11, S11], ..., [A1n, S1n]], ....,  [[Am1, S11], ..., [Amn, Smn]]  ]
% where:  Aij is a list of antecedents of a sequent (to the extent of this implementation, is always empty)
%		  Sij is a list of succeedents of a sequent
apply_invertible_to_premises(P0, P0):-
	are_premises_atomic(P0).
	
apply_invertible_to_premises(P0, P2) :-
	\+ are_premises_atomic(P0),
	apply_invertible_to_each_premise(P0, P1),
	apply_invertible_to_premises(P1, P2).

%% apply_invertible_to_each_premise(+Premises, -Premises)
%
% Iterates through the list of premises and applies the invertible rules to each sequent. 
apply_invertible_to_each_premise([], []).
	
apply_invertible_to_each_premise([H|T], P3):-
	apply_invertible_to_each_premise(T, P2),
	apply_invertible_to_each_premise_sequence(H, P1),
	append(P1, P2, P3).

%% apply_invertible_to_each_premise_sequence(+Premise, -Premise)
%
% Iteratets through the list of sequents of a premise and applies 
% the invertible rules. 
apply_invertible_to_each_premise_sequence([], []).

% covers the case: |- D, a + b		==> 	|- D, a, b
apply_invertible_to_each_premise_sequence([[Antecedent, Succedent]|T], P3) :-
		member(Ax1 + Ax2, Succedent),
		\+ number(Ax2),
		apply_invertible_to_each_premise_sequence(T, P1),
		remove(Ax1 + Ax2, Succedent, S1),
		append([Ax1, Ax2], S1, S2),
		append([[[Antecedent, S2]]], P1, P3).

% covers the case: |- D, a, 0		==> 	|- D, a		
apply_invertible_to_each_premise_sequence([[Antecedent, Succedent]|T], P3) :-
		member(0, Succedent),
		apply_invertible_to_each_premise_sequence(T, P1),
		remove(0, Succedent, S1),
		append([], S1, S2),
		append([[[Antecedent, S2]]], P1, P3).
		
% covers the case when one of the succeedents is the constant '1'.
apply_invertible_to_each_premise_sequence([[_, Succedent]|T], P1) :-
		member(1, Succedent),
		apply_invertible_to_each_premise_sequence(T, P1).

% covers the case: |- D, a & b		==> 	|- D, a		|- D, b
apply_invertible_to_each_premise_sequence([[Antecedent, Succedent]|T], P4):- 
		member(Ax1 & Ax2, Succedent),
		apply_invertible_to_each_premise_sequence(T, P1),
		remove(Ax1 & Ax2, Succedent, S1),
		append([Ax1], S1, S2),
		append([Ax2], S1, S3),
		append([[[Antecedent, S2]]], P1, P3),
		append([[[Antecedent, S3]]], P3, P4).

apply_invertible_to_each_premise_sequence([[Antecedent, Succedent]|T], P4):- 
		apply_invertible_to_each_premise_sequence(T, P1),
		append([[[Antecedent, Succedent]]], P1, P4).

%% are_premises_atomic(+Premises)
%
% checks whether the premises  contain any logical connectives or not.
% 
% +Premises 	- list of premies to be checked.
are_premises_atomic([]).
are_premises_atomic([H|T]) :-
	are_premises_atomic(T),
	is_premise_atomic(H).

%% is_premise_atomic(+Premise)
%
% checks whether a single premise contains a logicial symbol
is_premise_atomic([]).
is_premise_atomic([H|T]) :-
	is_premise_atomic(T),
	is_premise_sequences_atomic(H).
	

%% is_premise_sequences_atomic(+Sequent)
%
% checks whether a sequent of a premise contains a logical symbol
is_premise_sequences_atomic([_, Succedent]) :-
	\+ member(( _ & _ ), Succedent),
	\+ has_addition(Succedent), 
	\+ member(1, Succedent), 
	\+ member(0, Succedent).

%% has_addition(+Succedent)
%
% checks whether the succedent of a sequent contains a multiplicative or
has_addition(S) :- 
	member('Y' + X + Z, S), 
	number(X), atom(Z).
	
has_addition(S) :- 
	member('Y' + X + Z, S), 
	number(X), compound(Z).
	
has_addition(S) :- 
	member(X + Z, S), 
	atom(X), atom(Z).
	
has_addition(S) :- 
	member(X + Z, S), 
	X \= 'Y'+_,
	compound(X), atom(Z).
	
has_addition(S) :- 
	member(X + Z, S), 
	compound(X), compound(Z).	